#include <stdio.h>
#include <string.h>     // memcpy()
#include <stdint.h>     // uintptr_t
#include <assert.h>
#include <ctype.h>      // isspace()
#include "iter.h"
#include "gc.h"


#define not_reached() do { \
    fprintf(stderr, "%s:%d: assertion failed: code reached\n", \
                    __FILE__, __LINE__); \
    exit(1); \
} while(0)


// tests

// read characters from string that satisfy a predicate
struct Env_takewhile {
    int (*pred)(int);
    char *word;
};
static Iteratee f_takewhile_(void *env_, Stream *s)
{
    struct Env_takewhile *env = env_;
    char *word = env->word;
    
    if(s->state == END)
        return done(word);
    assert(s->state == CHUNK);

    if(s->n == 0)
        return cont(f_takewhile_, env);

    // count word characters in chunk
    const char *buf = s->elems;
    size_t i=0;
    while(i<s->n && env->pred(buf[i]))
        i++;

    if(i == 0) // word ended immediately
        return done(word);

    // append to word
    if(word) {
        size_t n = strlen(word);
        word = NEWA(char, n+i+1);
        memcpy(word, env->word, n);
        memcpy(word+n, buf, i);
        word[n+i] = '\0';
    }
    s->n -= i;
    s->elems += i;

    if(s->n > 0) // word ended inside chunk
        return done(word);

    // update environment and wait for more input
    env = DUP(env, struct Env_takewhile);
    env->word = word;
    return cont(f_takewhile_, env);
}
static Iteratee f_takewhile(void *env, Stream *s)
{
    if(s->state == END)
        return done("");

    if(s->elemsize != 1)
        return fail("input not a character stream");

    return f_takewhile_(env, s);
}
static Iteratee takewhile(int (*pred)(int))
{
    struct Env_takewhile *env = NEW(struct Env_takewhile);
    env->pred = pred;
    env->word = NEWA(char, 1);
    env->word[0] = '\0';
    return cont(f_takewhile, env);
}
static Iteratee dropwhile(int (*pred)(int))
{
    struct Env_takewhile *env = NEW(struct Env_takewhile);
    env->pred = pred;
    env->word = NULL;   // don't save result
    return cont(f_takewhile, env);
}

// a variation of takewhile that consumes at least one char
static Iteratee f_takewhile1(void *env_, Stream *s)
{
    struct Env_takewhile *env = env_;

    switch(s->state) {
    case END:
        return stop("EOF", f_takewhile1, env);
    case CHUNK:
        if(s->n < 1)
            return cont(f_takewhile1, env);
        else if(env->pred(((char *)s->elems)[0]))
            return f_takewhile(env, s);
        else
            return stop("no match", f_takewhile1, env);
    default:
        not_reached();
    }
}
Iteratee takewhile1(int (*pred)(int))
{
    Iteratee iter = takewhile(pred);
    assert(iter.state == CONT);
    iter.cont = f_takewhile1;
    return iter;
}
Iteratee dropwhile1(int (*pred)(int))
{
    Iteratee iter = dropwhile(pred);
    assert(iter.state == CONT);
    iter.cont = f_takewhile1;
    return iter;
}

// read a single word from the stream
static int iswordchar(int c) {
    return !isspace(c);
}
static Iteratee f_word(void *env, Stream *s) {
    Iteratee iter = takewhile1(iswordchar);
    assert(iter.state == CONT);
    return iter.cont(iter.env, s);
}
const Iteratee word = {.state = CONT, .cont = f_word, .env = NULL};

// drop leading word or whitespace from the stream
static Iteratee f_dropws(void *env, Stream *s) {
    int invert = (uintptr_t)env; // 0 = skip while ws; 1 = skip while not ws

    assert(s);
    if(s->state == END)
        return done(NULL);

    assert(s->elemsize == sizeof(char));
    int stop = 0;
    while(s->n && !stop) {
        stop = isspace(*(char *)s->elems);
        if(invert)
            stop = !stop;
        if(stop)
            break;

        s->elems += sizeof(char);
        s->n--;
    }

    if(s->n == 0)
        return cont(f_dropws, env);  // wait for more input
    else
        return done(NULL);
}
static Iteratee f_dropws1(void *env, Stream *s) {
    assert(s);
    if(s->state == END)
        return fail("dropws1: end of input");

    return f_dropws(env, s);
}
const Iteratee dropws = {.state = CONT, .cont = f_dropws, .env = (void *)0};
const Iteratee dropword = {.state = CONT, .cont = f_dropws1, .env = (void *)1};

// take leading whitespace from the stream
static Iteratee f_whitespace(void *env, Stream *s) {
    Iteratee iter = takewhile(isspace);
    assert(iter.state == CONT);
    return iter.cont(iter.env, s);
}
const Iteratee whitespace = {.state = CONT, .cont = f_whitespace, .env = NULL};

// read a single line from the stream
static int isnewline(int c) {
    return (c=='\n' || c=='\r');
}
static int islinechar(int c) {
    return !isnewline(c);
}
static Iteratee f_line2(void *env, void *result) {
    return seq_(2, dropwhile(isnewline), done(result));
}
static Iteratee f_line(void *env, Stream *s) {
    Iteratee iter = bind(takewhile(islinechar), f_line2, NULL);
    assert(iter.state == CONT);
    return iter.cont(iter.env, s);
}
const Iteratee line = {.state = CONT, .cont = f_line, .env = NULL};

// read one utf8-encoded unicode code point from the stream
//
//    0000 0000-0000 007F | 0xxxxxxx
//    0000 0080-0000 07FF | 110xxxxx 10xxxxxx
//    0000 0800-0000 FFFF | 1110xxxx 10xxxxxx 10xxxxxx
//    0001 0000-0010 FFFF | 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//
struct Env_utf8 {
    uint16_t cp;    // partially assembled code point
    uint8_t n;      // number of bytes remaining (1-3)
};
static Iteratee f_utf8_(void *env_, Stream *s) {
    struct Env_utf8 *env = env_;
    uint16_t cp;
    uint8_t b, n;

    assert(s->elemsize == 1);
    if(s->state == END)
        return stop("EOF", f_utf8_, env);
    assert(s->state == CHUNK);
    if(s->n == 0)
        return cont(f_utf8_, env);

    if(env) {
        // continue a partial decode
        // recover the loop state from env
        cp = env->cp;
        n  = env->n;
    } else {
        // start decoding a new character; read first byte
        b = ((uint8_t *)s->elems)[0];
        s->elems++; s->n--;

        // initialize loop state
        if(b>>7 == 0)       {n=0; cp = b;}
        else if(b>>5 == 6)  {n=1; cp = b & 0x1F;}
        else if(b>>4 == 14) {n=2; cp = b & 0x0F;}
        else if(b>>3 == 30) {n=3; cp = b & 0x07;}
        else
            return stop("invalid utf8 (start)", f_utf8_, NULL);
            // (resume = retry on next byte)
    }

    // read remaining bytes and integrate them into cp
    for(; n>0; n--) {
        if(s->n == 0) {
            // end of chunk, suspend
            env = NEW(struct Env_utf8);
            env->cp = cp;
            env->n = n;
            return cont(f_utf8_, env);
        }

        // read next byte
        b = *((uint8_t *)s->elems);

        if(b>>6 != 2) {
            env = NEW(struct Env_utf8);
            env->cp = cp;
            env->n = n;
            return stop("invalid utf8 (continuation)", f_utf8_, NULL);
            // (resume = restart on this byte)
        }

        // byte is valid input; consume it
        s->elems++; s->n--;

        cp = (cp << 6) | (b & 0x3F);
    }

    // XXX surrogate pairs

    return done((void *)(uintptr_t)cp);
}
static Iteratee f_utf8(void *env, Stream *s)
{
    if(s->elemsize != 1)
        return fail("input not a byte stream");
    return f_utf8_(env, s);
}
const Iteratee utf8 = {.state = CONT, .cont = f_utf8, .env = NULL};

int uisspace(void *elem)
{
    uintptr_t cp = *(uintptr_t *)elem;  // Unicode code point
    return isspace((int)cp);            // XXX recognizes ASCII only for now
}

// take a word from a stream of unicode code points
static Iteratee f_uword(void *env, Stream *s)
{
    Iteratee dropwhile_isspace = wrap(ewhile(uisspace), drop);
    Iteratee takeuntil_isspace = wrap(until(uisspace), take1);
    Iteratee iter = seq_(2, dropwhile_isspace, takeuntil_isspace);
    assert(iter.state == CONT);
    return iter.cont(iter.env, s);
}
const Iteratee uword = {.state = CONT, .cont = f_uword, .env = NULL};

// drop a word from a stream of unicode code points
static Iteratee f_uword_(void *env, Stream *s)
{
    Iteratee dropwhile_isspace = wrap(ewhile(uisspace), drop);
    Iteratee dropuntil_isspace = wrap(until(uisspace), drop);
    Iteratee iter = seq_(2, dropwhile_isspace, dropuntil_isspace);
    assert(iter.state == CONT);
    return iter.cont(iter.env, s);
}
const Iteratee uword_ = {.state = CONT, .cont = f_uword_, .env = NULL};


// TODO: print each word from stdin

// extract the first utf-8-encoded word from stdin
int test8(int argc, char **argv)
{
    Iteratee ie = apply(enumf(stdin), wrap(decode(utf8), uword));
    Stream *s = finish(ie);
    printf("%zu\n", s? s->n:0); // XXX actually print the word
    return 0;
}

// count utf-8-encoded words on stdin
int test7(int argc, char **argv)
{
    Iteratee wcount = wrap(decode(uword_), count);
    Iteratee uwcount = wrap(decode(utf8), wcount);

    Iteratee i2 = apply(enumf(stdin), uwcount);
    uintptr_t count = (uintptr_t)finish(i2);

    printf("%zu\n", count);
    return 0;
}

// count the number of utf-8 characters on stdin
int test6(int argc, char **argv)
{
    Iteratee outer = apply(enumf(stdin), apply(decode(utf8), count));
        // outer = the iterated 'ucount'
    Iteratee *inner = finish(outer);
        // inner = the iterated 'count'
    if(!inner) {
        // try to get partial result
        if(outer.state == STOP && outer.cont) {
            fprintf(stderr, "try to resume\n");
            outer.state = CONT; // resume
            inner = finish(outer);
        }
        if(!inner)
            return 1;   // unrecoverable decoder error
    }
    uintptr_t count = (uintptr_t)finish(*inner);

    printf("%zu\n", count);
    return 0;
}

// count lines on stdin
int test5(int argc, char **argv)
{
    Iteratee countlines = apply(decode(line), count);

    Iteratee outer = apply(enumf(stdin), countlines);
        // outer = the iterated 'countlines'
    Iteratee *inner = finish(outer);
        // inner = the iterated 'count'
    if(!inner) {
        // try to get partial result
        if(outer.state == STOP && outer.cont) {
            fprintf(stderr, "try to resume\n");
            outer.state = CONT; // resume
            inner = finish(outer);
        }
        if(!inner)
            return 1;   // unrecoverable decoder error
    }
    uintptr_t count = (uintptr_t)finish(*inner);

    printf("%zu\n", count);
    return 0;
}

// count words on stdin
int test4(int argc, char **argv)
{
    Enumeratee words = decode(bind_(dropws, dropword));
    Iteratee countwords = apply(words, count);

    Iteratee outer = apply(enumf(stdin), countwords);
        // outer = the iterated 'countwords'
    Iteratee *inner = finish(outer);
        // inner = the iterated 'count'
    if(!inner) {
        // try to get partial result
        if(outer.state == STOP && outer.cont) {
            fprintf(stderr, "try to resume\n");
            outer.state = CONT; // resume
            inner = finish(outer);
        }
        if(!inner)
            return 1;   // unrecoverable decoder error
    }
    uintptr_t count = (uintptr_t)finish(*inner);

    printf("%zu\n", count);
    return 0;
}

// extract the first word from stdin
int test3(int argc, char **argv)
{
    const char *w = finish(apply(enumf(stdin), bind_(dropws, word)));
    printf("%s\n", w? w:"(null)");
    return 0;
}

// count the number of bytes in all files on the cmdline
int test2(int argc, char **argv)
{
    Enumerator en;

    if(argc < 2) {
        en = enumf(stdin);
    } else {
        en = enumfile(argv[1]);
        for(int i=2; i<argc; i++)
            en = append(en, enumfile(argv[i]));
    }

    void *result = finish(apply(en, count));
    uintptr_t count = (uintptr_t)result;

    printf("%zu\n", count);
    return 0;
}

// count the number of bytes passed on stdin
int test1(int argc, char **argv)
{
    void *result = finish(apply(enumf(stdin), count));
    uintptr_t count = (uintptr_t)result;

    printf("%zu\n", count);
    return 0;
}

int main(int argc, char **argv)
{
    gc_init(argv);

    int r = test4(argc, argv);

    gc_stats(stderr);
    return r;
}
